package com.kms.service;

import com.j256.ormlite.dao.Dao;
import com.kms.crosscut.SoapFaultHelper;
import com.kms.dto.payment.*;
import com.kms.model.Payment;
import lombok.RequiredArgsConstructor;

import javax.jws.HandlerChain;
import javax.jws.WebMethod;
import javax.jws.WebParam;
import javax.jws.WebService;
import java.sql.SQLException;
import java.sql.Timestamp;
import java.util.List;
import java.util.Objects;
import java.util.stream.Collectors;

@WebService
@HandlerChain(file = "handlers.xml")
@RequiredArgsConstructor
public class PaymentService {

    private final Dao<Payment,Integer> paymentDao;

    @WebMethod
    public PaymentInitResp initPayment(@WebParam(name = "req") PaymentInitReq req) {
        verifyInitPaymentData(req);
        try {
            List<Payment> idempotentCheck = paymentDao.queryForMatching(Payment.builder()
                    .initiatorId(req.getInitiatorId())
                    .idempotentId(req.getIdempotentId()).build());

            if (!idempotentCheck.isEmpty()) {
                Payment payment = idempotentCheck.get(0);
                if (payment.getAmount() != req.getAmount() || payment.isPaid()) {
                   return new PaymentInitResp(null, false, 3,
                           "You're trying to create another payment request with duplicate idempotency key");
                }
                return new PaymentInitResp(payment.getPaymentId(), true, 1, "");
            }

            Payment payment = Payment.builder()
                    .amount(req.getAmount())
                    .paymentInitTime(new Timestamp(System.currentTimeMillis()))
                    .initiatorId(req.getInitiatorId())
                    .idempotentId(req.getIdempotentId())
                    .description(req.getDescription())
                    .build();

            paymentDao.create(payment);
            return new PaymentInitResp(payment.getPaymentId(), true, 1, "");
        } catch (SQLException e) {
            return new PaymentInitResp(null, false, 2,
                    "Unexpected error has happened");
        }
    }

    @WebMethod
    public PaymentStatusResp checkPaymentStatus(@WebParam(name = "paymentId") Integer paymentId) {
        try {
            Payment payment = paymentDao.queryForId(paymentId);

            if (payment == null) {
                return new PaymentStatusResp(null, "Payment id not found");
            }
            return new PaymentStatusResp(payment.isPaid(), null);
        } catch (SQLException e) {
            return new PaymentStatusResp(null, "Unexpected error has happened");
        }
    }

    @WebMethod
    public PaymentStatusBatchResp checkPaymentStatusBatch(@WebParam(name = "paymentIds") List<Integer> paymentIds) {
        List<Payment> payments = paymentIds.stream().map(paymentId -> {
                    try {
                        return paymentDao.queryBuilder()
                                .orderBy("payment_id", false)
                                .where().eq("payment_id", paymentId)
                                .queryForFirst();
                    } catch (SQLException e) {
                        return null;
                    }
                }).filter(Objects::nonNull)
                .collect(Collectors.toList());
        if (payments.isEmpty()) {
            return new PaymentStatusBatchResp("Payment ids not found or unexpected error has happened", null);
        }
        return new PaymentStatusBatchResp(null, payments.stream()
                .map(payment -> PaymentStatusBatchResp.PaymentWrapper.builder()
                            .paymentId(payment.getPaymentId())
                            .paid(payment.isPaid())
                            .initiatorId(Integer.valueOf(payment.getInitiatorId()))
                            .build()).toArray(PaymentStatusBatchResp.PaymentWrapper[]::new));
    }

    private void verifyInitPaymentData(PaymentInitReq req) {
        if (req == null) {
            SoapFaultHelper.throwFault("InvalidRequest", "Request is empty");
        }
        if (req.getInitiatorId() == null) {
            SoapFaultHelper.throwFault("InvalidRequest", "Payment initiator id is empty");
        }
        if (req.getIdempotentId() == null) {
            SoapFaultHelper.throwFault("InvalidRequest", "Payment idempotent id is empty");
        }
        if (req.getAmount() <= 0) {
            SoapFaultHelper.throwFault("InvalidRequest", "Payment amount is invalid");
        }
    }

}
